#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dir.h>

/*
 *  Klass, mis tegeleb etteantud faili andmete bitthaaval kirjutamisega,
 *  puhverdades need kigepealt mllu. Lpus eemaldatakse faili lpust ka
 *  kik nullid, sest selle arvelt saab ruumi kokku hoida. Kui faili lpp
 *  vastu tuleb, hakkab biti lugemise operatsioon nulle tagastama.
 */ 
#define BITFILE_BUFFERLENGTH 1000
#define BITFILE_CLOSED 0
#define BITFILE_READ 1
#define BITFILE_WRITE 2

class BitFile {
  public:
    BitFile ();
    ~BitFile ();
    bool OpenForReading (char* filename); // Ava fail lugemiseks
    bool OpenForWriting (char* filename); // Ava fail kirjutamiseks
    void Close ();                        // Sulge fail
    bool EndReached ();                   // Kas fail on otsas?
    bool ReadBit ();                      // Loe bitt
    void WriteBit (bool bit);             // Kirjuta bitt

  protected:
    void Flush ();                        // Kirjutab puhvri kettale
    void Truncate ();                     // Eemaldab lpust nullid

    char* tempname;                       // Ajutise faili nimi
    char* name;                           // Faili nimi
    FILE* bitfile;                        // Failisang
    unsigned char* bytedata;              // Mlupuhver
    int bytes;                            // Puhvri indeks
    int bytesread;                        // Loetud baite kettalt
    int bitnum;                           // Aktiivne bitt
    int mode;                             // Faili olek
    int bitmask;                          // Bitimask
    int zerooffset;                       // Nullide algusaadress
    int offset;                           // Jooksev failiaadress
};

BitFile::BitFile ()
  : bitfile (NULL),
    tempname (NULL),
    name (NULL),
    bytedata (NULL),
    bytes (0),
    bitnum (0),
    bytesread (0),
    bitmask (1),
    zerooffset (-1),
    offset (0),
    mode (BITFILE_CLOSED)
{
  bytedata = (unsigned char*)malloc (BITFILE_BUFFERLENGTH);
  if (bytedata == NULL) {
    printf ("BitFile: viga mlu haaramisel\n");
  } else {
    memset (bytedata, 0, BITFILE_BUFFERLENGTH);
  }
}

BitFile::~BitFile () {
  Close ();
  if (bytedata != NULL)
    free (bytedata);
  if (name != NULL)
    free (name);
  if (tempname != NULL)
    free (tempname);
  bytedata = 0;
}

bool BitFile::OpenForReading (char* filename) {
  Close ();
  bitfile = fopen (filename, "rb");
  if (bitfile != NULL) {
    mode = BITFILE_READ;
    return 1;
  } else
    return 0;
}

bool BitFile::OpenForWriting (char* filename) {
  Close ();
  name = strdup (filename);
  tempname = tmpnam (NULL); 
  bitfile = fopen (tempname, "wb");
  if (bitfile != NULL) {
    mode = BITFILE_WRITE;
    return 1;
  } else
    return 0;
}

void BitFile::Flush () {
  if (mode == BITFILE_WRITE) {
    if (bitnum == 0)
      fwrite(bytedata, 1, bytes, bitfile);
    else
      fwrite(bytedata, 1, bytes + 1, bitfile);
    bytes = 0;
  }
}

void BitFile::Truncate () {
  if (mode == BITFILE_WRITE) {
    if (zerooffset == -1)
      zerooffset = offset + 1;
    FILE* i = fopen (tempname, "rb");
    FILE* o = fopen (name, "wb");
    if (i == NULL || o == NULL) {
      printf ("BitFile: Truncation failed!");
      return;
    }
    int bytesleft = zerooffset;
    do {
      bytesread = fread(bytedata, 1, BITFILE_BUFFERLENGTH, i); 
      if (bytesread <= bytesleft) {
        bytesleft -= bytesread;
      } else {
        bytesread = bytesleft;
        bytesleft = 0;
      }
      if (bytesread > 0)
        fwrite(bytedata, 1, bytesread, o);
    } while (bytesleft != 0);
    fclose (i);
    fclose (o);
    unlink (tempname);
    free (tempname);
    free (name);
  }
}

void BitFile::Close () {
  Flush ();
  if (bitfile != NULL) {
    fclose (bitfile);
  }
  Truncate ();
  bitnum = 0;
  bitmask = 1;
}

bool BitFile::EndReached () {
  if (feof (bitfile) && bytes == bytesread)
    return 1;
  else
    return 0;
}

bool BitFile::ReadBit () {
  if (EndReached ())
    return 0;

  if (bytes == bytesread) {
    memset (bytedata, 0, BITFILE_BUFFERLENGTH);
    bytesread = fread(bytedata, 1, BITFILE_BUFFERLENGTH, bitfile); 
    if (bytesread == 0) {
      return 0;
    }
    bitnum = 0;
    bytes = 0;
    bitmask = 1;
  }
  bool bit = bytedata[bytes] & bitmask;
  bitnum++;
  bitmask <<= 1;
  if (bitnum == 8) {
    bytes++;
    bitnum = 0;
    bitmask = 1;
  }

  return bit;
}

void BitFile::WriteBit (bool bit) {
  if (bytes == BITFILE_BUFFERLENGTH) {
    if (fwrite(bytedata, 1, bytes, bitfile) == bytes) {
      memset (bytedata, 0, BITFILE_BUFFERLENGTH);
      bytes = 0;
      bitnum = 0;
      bitmask = 1;
    } else {
      printf ("BitFile: viga faili kirjutamisel!\n");
    }
  }
  if (bit == 1) {
    bytedata[bytes] += bitmask;
  }
  bitnum++;
  bitmask <<= 1;
  if (bitnum == 8) {
    if (bytedata[bytes] == 0) {
      if (zerooffset <= 0)
          zerooffset = offset;
    } else {
      zerooffset = -1;
    }
    bytes++;
    offset++;
    bitmask = 1;
    bitnum = 0;
  }
}

/*
 * Kahendpuu tippu esindav klass
 */
class Node {
  public:
    Node ();
    Node (Node* l, Node* r);
    Node* leftchild;                      // Viit vasakule ttarkolmnurgale   
    Node* rightchild;                     // Viit paremale ttarkolmnurgale
};

Node::Node () {
  leftchild = NULL;
  rightchild = NULL;
}

Node::Node (Node* l, Node* r) {
  leftchild = l;
  rightchild = r;
}

/*
 * Puuoperatsioone (failist lugemist ja faili kirjutamist ning ekraanile 
 * vljastust) sisaldav klass. Samuti on lisatud meetod pseudo-juhusliku
 * kolmnurkade kahendpuu genereerimiseks.
 */

class TreeFile {
  public:
    TreeFile ();
    ~TreeFile ();
    Node* ReadTree (char* filename);              // Failist lugemine
    void WriteTree (char* filename, Node* root);  // Faili salvestamine
    void BitPrint (Node* node); // Puu bittkuju vljastamine
    void GenTree (Node* root);  // Pseude-juhusliku puu genereerimine

  protected:
    void WriteNode (Node* root);
    Node* ReadNode ();
    BitFile* file;
    int limit;                  // Genereeritava puu tippude piirarv
};

TreeFile::TreeFile () : file (NULL), limit (50) {}
TreeFile::~TreeFile () {}

void TreeFile::WriteTree (char* filename, Node* root) {
  file = new BitFile ();
  if (!file->OpenForWriting (filename)) {
    printf ("TreeFile: viga faili '%s' lugemiseks avamisel!\n", filename);
    return;
  }
  WriteNode (root);
  file->Close ();
}

void TreeFile::WriteNode (Node* node) {
  if (node->leftchild != NULL && node->rightchild != NULL) {
    file->WriteBit (1);
    WriteNode (node->leftchild);
    WriteNode (node->rightchild);
  } else {
    file->WriteBit (0);
  }
}

Node* TreeFile::ReadTree (char* filename) {
  file = new BitFile ();
  if (!file->OpenForReading (filename)) {
    printf ("TreeFile: viga faili '%s' kirjutamiseks avamisel!\n", filename);
    return NULL;
  }
  Node* root = ReadNode ();
  file->Close ();
  return root;
}

void TreeFile::GenTree (Node* node) {
  if (limit <= 0)
    return;
  int a = rand () % 100;
  if (a < 80) {
    limit -= 2;
    node->leftchild = new Node ();
    GenTree (node->leftchild);
    node->rightchild = new Node ();
    GenTree (node->rightchild);
  }
}

Node* TreeFile::ReadNode () {
  Node* node = new Node ();
  bool split = file->ReadBit ();
  if (split) {
    node->leftchild = ReadNode ();
    node->rightchild = ReadNode ();
  }
  return node;
}

void TreeFile::BitPrint (Node* node) {
  if (node->leftchild != NULL && node->rightchild != NULL) {
    printf ("1");
    BitPrint (node->leftchild);
    BitPrint (node->rightchild);
  } else {
    printf ("0");
  }
}

/*
 * Phifunktsioon
 */

int main(int argc, char *argv[])
{
  srand (time (NULL));
  // Loome he kahendpuu ja kujutame seda ekraanil
  printf ("Koostan kolmnurkade kahendpuu puu...\n");
  Node* root = new Node (new Node (new Node (new Node (), new Node ()), new Node ()), new Node (new Node (), new Node (new Node (), new Node ())));
  TreeFile* tf = new TreeFile ();
  printf ("Loodud puu bittkuju: ");
  tf->BitPrint (root);
  printf ("\n");
  printf ("Salvestan puu faili 'test.tre'...\n");
  tf->WriteTree ("test.tre", root);
  printf ("Loen puu failist 'test.tre'...\n");
  Node* reload = tf->ReadTree ("test.tre");
  printf ("Loetud puu bittkuju: ");
  tf->BitPrint (reload);
  printf ("\n");

  printf ("Genereerin suvalise kolmnurkade kahendpuu...\n");
  Node* random = new Node ();
  tf->GenTree (random);
  printf ("Saadud puu bittkuju: ");
  tf->BitPrint (random);
  printf ("\n");
  printf ("Salvestan puu faili 'suvaline.tre'...\n");
  tf->WriteTree ("suvaline.tre", random);
  printf ("Loen puu failist 'suvaline.tre'...\n");
  Node* randomreload = tf->ReadTree ("suvaline.tre");
  printf ("Loetud puu bittkuju: ");
  tf->BitPrint (randomreload);
  printf ("\n");
  delete tf;
  return 0;
}